import unittest
from scibatt.readers import biologic
import os


class TestBioLogic(unittest.TestCase):
    def test_read_txt_with_valid_datafile(self):
        data_file_path = os.path.join(
            os.path.dirname(__file__), "test_data", "biologic.mpr"
        )
        data = biologic.read_mpr(data_file_path)

        self.assertIsInstance(data, dict)


if __name__ == "__main__":
    unittest.main()



def select_halfcycles(data, number, select_range=None):
    """
    Returns input data with only the selected halfcycle numbers.
    number can be integer og list of integers.
    Range is a tuple of integers, or list of tuples of integers.
    """

    selected = {}
    halfcycles_to_catch = []
    if isinstance(number, int):
        halfcycles_to_catch.append(number)
    elif isinstance(number, list):
        for num in number:
            if num not in halfcycles_to_catch:
                halfcycles_to_catch.append(num)

    if isinstance(select_range, tuple):
        for num in range(
            select_range[0], select_range[1] + 1
        ):  # Including last element
            if num not in halfcycles_to_catch:
                halfcycles_to_catch.append(num)
    elif isinstance(select_range, list):
        for selected_range in select_range:
            for num in range(
                selected_range[0], selected_range[1] + 1
            ):  # Including last element
                if num not in halfcycles_to_catch:
                    halfcycles_to_catch.append(num)

    for i, (key, value) in enumerate(data.items()):
        if i + 1 in halfcycles_to_catch:
            selected[key] = value
    return selected



import matplotlib.pyplot as plt
import numpy as np


class GalvanoStatic:
    def __init__(self, title="Galvanostatic", style=None):
        if style:
            if style == "Evyon":
                from eir_visualize import EvyonStyle

                self.fig, self.ax = EvyonStyle().__enter__()
        else:
            self.fig, self.ax = plt.subplots(nrows=1, ncols=1)
        # self.fig.suptitle(suptitle) # This is figure title, not axis
        self.ax.set(
            title=title,
            # ylabel = 'Potential [V]',
            # xlabel = 'Specific Capacity [mAh]',
            # ylim = (2.5,5),
            # xlim = (0, 150),
            # xticks = (np.arange(0, 150), step=20)),
            # yticks = (np.arange(3, 5, step=0.2)),
        )
        self.fig.tight_layout()
        self.twinx = None

    def get_twinx(self):
        if not self.twinx:
            self.twinx = self.ax.twinx()
            return self.twinx
        else:
            return self.twinx

    def add_data(self, x, y1, y2=None, labely1=None, labely2=None, vistype="scatter"):
        dotsize = 2
        # Plot data
        if vistype == "scatter":
            self.ax.scatter(x, y1, label=labely1 if labely1 else "", s=dotsize)
        else:
            self.ax.plot(x, y1, label=labely1 if labely1 else "")
        if type(y2) == np.ndarray:
            if vistype == "scatter":
                self.get_twinx().scatter(
                    x, y2, label=labely2 if labely2 else "", s=dotsize, alpha=0.5
                )
            else:
                self.get_twinx().plot(
                    x, y2, label=labely2 if labely2 else "", linestyle="--"
                )

    def set_x(self, xlabel="Specific Capacity [mAh]"):
        self.ax.set_xlabel(xlabel)

    def set_y(self, ylabel="Potential, U [V]"):
        self.ax.set_ylabel(ylabel)

    def set_y2(self, ylabel="Current, I [A]"):
        self.get_twinx().set_ylabel(ylabel)

    def save(self, filename=None, format="svg"):
        format = format.strip(".")
        if filename and not filename.endswith(f".{format}"):
            filename += "." + format
        plt.savefig(
            filename if filename else f"plot.{format}",
            dpi=300,
            bbox_inches="tight",
            format=format,
        )

    def plot(self):
        plt.legend()
        plt.show()



import pandas as pd
from scibatt.config import COLUMN_NAMES


def transform_timebased(data, exp_time=True):
    x = []
    y = []
    for key, value in data.items():
        x.append(
            (
                value[COLUMN_NAMES["TIME"]] - value[COLUMN_NAMES["TIME"]].iloc[0]
            ).to_numpy()
        )
        y.append(value[COLUMN_NAMES["VOLTAGE1"]].to_numpy())
    return (x, y)


def transform_capacitybased(data, unit="As"):
    x = []
    y = []
    z = []
    for key, df in data.items():
        # Convert to experiment time
        df[COLUMN_NAMES["TIME"]] = (
            df[COLUMN_NAMES["TIME"]] - df[COLUMN_NAMES["TIME"]].iloc[0]
        )

        # Calculate the change in COLUMN_NAME_TIME
        df["dt"] = df[COLUMN_NAMES["TIME"]].diff()

        # Calculate the cumulative product of "I" and "dt"
        df[COLUMN_NAMES["CHARGE"]] = (
            df[COLUMN_NAMES["CURRENT"]] * df["dt"].fillna(0).cumsum()
        )

        if unit == "mAh":
            df[COLUMN_NAMES["CHARGE"]] = df[COLUMN_NAMES["CHARGE"]] * 1000 / 60 / 60

        x.append(df[COLUMN_NAMES["CHARGE"]].to_numpy())
        y.append(df[COLUMN_NAMES["VOLTAGE1"]].to_numpy())
        z.append(df[COLUMN_NAMES["CURRENT"]].to_numpy())
    return (x, y, z)


def transform_raw(data):
    dfs = []
    for key, df in data.items():
        dfs.append(df)
    concat = pd.concat(dfs)
    return (
        concat[COLUMN_NAMES["TIME"]].to_numpy(),
        concat[COLUMN_NAMES["VOLTAGE1"]].to_numpy(),
        concat[COLUMN_NAMES["CURRENT"]].to_numpy(),
    )



import unittest
from scibatt.readers import maccor
import os


class TestMaccor(unittest.TestCase):
    def test_read_txt_with_valid_datafile(self):
        data_file_path = os.path.join(
            os.path.dirname(__file__), "test_data", "maccor.txt"
        )
        data = maccor.read_txt(data_file_path)

        self.assertIsInstance(data, dict)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Data-readers for BioLogic"""

from galvani import BioLogic
import pandas as pd
import datetime
from scibatt.config import COLUMN_NAMES, CURRENT_ZERO_TOLERANCE


def read_mpr(filepath):
    mpr_file = BioLogic.MPRfile(filepath)
    print(mpr_file.__dir__())

    df = pd.DataFrame(mpr_file.data)

    # Modifying time
    def convert_timestamp_to_unix_epoch(time, start_timestamp):
        datetime_obj = start_timestamp + datetime.timedelta(seconds=time)
        return datetime_obj.timestamp()  # Returns unix epoch float

    df[COLUMN_NAMES["TIME"]] = df["time/s"].apply(
        lambda x: convert_timestamp_to_unix_epoch(x, mpr_file.timestamp)
    )
    df[COLUMN_NAMES["CURRENT"]] = df["control/V/mA"].apply(
        lambda x: x / 1000
    )  # Convert to Amperes

    # Rename columns to match spec
    df.rename(
        columns={
            "Ewe/V": COLUMN_NAMES["VOLTAGE1"],
        },
        inplace=True,
    )

    # Group by step to separate steps
    groups_step = df.groupby("Ns")

    # Scan groups and add to return dict
    data = {}
    tol = CURRENT_ZERO_TOLERANCE
    for num, group_df in groups_step:
        mean_current = group_df[COLUMN_NAMES["CURRENT"]].mean()
        timestamp = group_df[COLUMN_NAMES["TIME"]].iloc[0]

        # Remove columns we don't want
        required_columns = [
            COLUMN_NAMES["TIME"],
            COLUMN_NAMES["CURRENT"],
            COLUMN_NAMES["VOLTAGE1"],
        ]
        group_df = group_df.drop(
            columns=[col for col in df if col not in required_columns]
        )

        if -tol < mean_current < tol:
            data[f"{timestamp}_cycling_p000.000A"] = group_df
        elif mean_current > tol:
            data[f"{timestamp}_cycling_p{mean_current:08.4f}A"] = group_df
        elif mean_current < tol:
            data[f"{timestamp}_cycling_n{mean_current:08.4f}A"] = group_df

    return data



# -*- coding: utf-8 -*-
"""Data-readers for Maccor"""

import pandas as pd
import datetime
from scibatt.config import COLUMN_NAMES, CURRENT_ZERO_TOLERANCE


def read_txt(filepath):
    """
    Reads a maccor datafile and returns a dict of of dataframes;
    The key being the standard filename, and the dataframe being the data for each step in the programme
    """
    # This reads the csv file with some extra options
    with open(filepath, "r") as f:
        header_line_num = None
        for i, line in enumerate(f.readlines()):
            if line.startswith("Rec\tCycle"):
                header_line_num = i
                break
        if not header_line_num:
            raise Exception(
                f"Could not find headerline in Maccor datafile: {filepath}"
            )  # TODO: Convert to custom exception

    df = pd.read_csv(
        filepath,
        usecols=["Cycle", "DPT Time", "Current", "Voltage", "Capacity", "Step", "MD"],
        encoding="UTF-8",
        header=header_line_num,
        delimiter="\t",
    )

    # Modifying time
    def convert_timestamp_to_unix_epoch(timestamp_str):
        timestamp_format = "%m/%d/%Y %H:%M:%S"
        datetime_obj = datetime.datetime.strptime(timestamp_str, timestamp_format)

        return datetime_obj.timestamp()  # Returns unix epoch float

    df[COLUMN_NAMES["TIME"]] = df["DPT Time"].apply(
        lambda x: convert_timestamp_to_unix_epoch(x)
    )

    # Making sure negative currents are negative
    df.loc[df["MD"] == "D", "Current"] *= -1

    # Rename columns to match spec
    df.rename(
        columns={
            "Current": COLUMN_NAMES["CURRENT"],
            "Voltage": COLUMN_NAMES["VOLTAGE1"],
            "Cycle": "Cycle",
        },
        inplace=True,
    )

    # Group by step to separate steps
    groups_step = df.groupby("Step")

    # Scan groups and add to return dict
    data = {}
    tol = CURRENT_ZERO_TOLERANCE
    for num, group_df in groups_step:
        mean_current = group_df[COLUMN_NAMES["CURRENT"]].mean()
        timestamp = group_df[COLUMN_NAMES["TIME"]].iloc[0]

        # Remove columns we don't want
        required_columns = [
            COLUMN_NAMES["TIME"],
            COLUMN_NAMES["CURRENT"],
            COLUMN_NAMES["VOLTAGE1"],
        ]
        group_df = group_df.drop(
            columns=[col for col in df if col not in required_columns]
        )

        if -tol < mean_current < tol:
            data[f"{timestamp}_cycling_p000.000A"] = group_df
        elif mean_current > tol:
            data[f"{timestamp}_cycling_p{mean_current:08.4f}A"] = group_df
        elif mean_current < tol:
            data[f"{timestamp}_cycling_n{mean_current:08.4f}A"] = group_df

    return data
